#include "bpp8.h"

#define BPP_COLOR 256

typedef struct {
  int r;
  int g;
  int b;
} PALETTE;

PALETTE PAL256[BPP_COLOR];
static int fore_index = 255;
static int back_index = 0;
int color_index = 0;

#define PAL PAL256

int bpp8_nearest(int r,int g,int b)
{
  int dist=0xffffff;
  int n,d,i,R,G,B;
  for (n=0,i=0;n<BPP_COLOR;n++) {
    R = r - PAL[n].r; R*=R;
    G = g - PAL[n].g; G*=G;
    B = b - PAL[n].b; B*=B;
    d = R + G + B;
    if (d<dist) { i = n; dist = d; }
  }
  return i;
}

int bpp8_setcolor(int r,int g,int b)
{
  color_index = bpp8_nearest(r,g,b);
  vga_setcolor(color_index);
}

extern int bpp8_bitblt(int x,int y,int w,int h,char *bmp);
extern int bpp8_bitblt_1024(int x,int y,int w,int h,char *bmp);
extern int bpp8_getbitmap(int x,int y,int w,int h,char *bmp);
extern int bpp8_getbitmap_1024(int x,int y,int w,int h,char *bmp);
extern int bpp8_scroll(int x,int y,int w,int h,int line);
extern int bpp8_scroll_1024(int x,int y,int w,int h,int line);
extern int bpp8_fillbox(int x,int y,int w,int h);
extern int bpp8_fillbox_1024(int x,int y,int w,int h);
extern int bpp8_xorbox(int x,int y,int w,int h);
extern int bpp8_xorbox_1024(int x,int y,int w,int h);

extern void bpp8_fnt2image(
       int length,
       int bits,
       unsigned char fg,
       unsigned char bg,
       unsigned char *fnt,
       unsigned char *image
     );

int bpp8_drawfont(int x,int y,int w,int h,unsigned char *bmp)
{
  int leng = w * h;
  unsigned char * FNT = (unsigned char *)malloc(sizeof(char)*leng);

  bpp8_fnt2image(leng,w,fore_index,back_index,bmp,FNT);
  bitblt(x,y,w,h,FNT);

  free(FNT);
}

int bpp8_setforeground(int r,int g,int b)
{
  set_swcolor(foreground,r,g,b);
  fore_index = bpp8_nearest(r,g,b);
  return fore_index;
}

int bpp8_setbackground(int r,int g,int b)
{
  set_swcolor(background,r,g,b);
  back_index = bpp8_nearest(r,g,b);
  return back_index;
}

int bpp8_miximage(unsigned char *dest,unsigned char *src,int w,int h)
{
  int n = w * h;
  while (n>0) {
    if (*src) *dest=*src;
    dest++; src++; n--;
  }
  return 1;
}

unsigned char * bpp8_xpm(int *w,int *h,char **xpm)
{
  int x,y;
  int W,H;
  int colors;
  int n,i;
  int present;
  unsigned char * image;
  PALETTE *pal;
  int * MAP;
  char ** KEY;
  char * buff;
  unsigned char *p;

  if (sscanf(xpm[0],"%d %d %d %d",w,h,&colors,&present)!=4) return NULL;
  W = *w; H = *h;

  /* allocate memory */
  image = (unsigned char *)malloc(W*H*sizeof(char));

  /* color */
  pal = (PALETTE*)malloc(sizeof(PALETTE)*colors);
  MAP = (int*)malloc(sizeof(int)*colors);
  KEY = (char**)malloc(sizeof(char*)*colors);

  for (i=0,n=1;i<colors;i++,n++) {
    KEY[i]=(char*)malloc(sizeof(char)*(present+1));
    KEY[i][present]=0;
    memcpy(KEY[i],xpm[n],present);
    buff = &xpm[n][present];
    while (*buff&&*buff!='c') buff++;
    if (*buff=='c') buff++;
    if (*buff==' ') buff++;
    if (strcasecmp(buff,"None")==0) {
      memset(&pal[i],0,sizeof(PALETTE)); 
    } else {
      if (*buff=='#') buff++;
      #define GETID(v) (v>='0'&&v<='9') ? v-'0': ((v>='A'&&v<='Z') ? v-'A'+10: ((v>='a'&&v<='z') ? v-'a'+10:0))
      #define GETCOLOR(c) c=GETID(*buff); c<<=4; buff++; c|=GETID(*buff); buff++
      GETCOLOR(pal[i].r);
      GETCOLOR(pal[i].g);
      GETCOLOR(pal[i].b);
      #undef  GETCOLOR
      #undef  GETID
    }
    MAP[i] = bpp8_nearest(pal[i].r,pal[i].g,pal[i].b);
  }
  p = image; *p = 0;
  for (y=0;y<H;y++,n++) {
    for (x=0;x<W;x++) {
      for (i=0;i<colors;i++)
        if (strncmp(KEY[i],&xpm[n][x*present],present)==0) break;
      *p = (char)MAP[i];
      p++;
    }
  }
  
  free(pal);
  free(MAP);
  for (i=0;i<colors;i++) free(KEY[i]);
  free(KEY);
  return image;
}

int bpp8_mouseinit(char **xpm)
{
  mouse_bitmap = create_image_xpm(&mouse_width,&mouse_height,xpm);
  mouse_memsize = mouse_width*mouse_height*sizeof(char);
  mouse_screen = (unsigned char *)malloc(mouse_memsize);
  mouse_paint  = (unsigned char *)malloc(mouse_memsize);
  
  return 1;
}

int bpp8_mouserecover(void)
{
  if (!mouse_enable) return 0;
  if (mouse_painted) {
    bitblt(mouse_x,mouse_y,mouse_width,mouse_height,mouse_screen);
  }
  mouse_painted = 0;
  mice_update  = 0;
  return 1;
}

int bpp8_mousemove(int x,int y)
{
  if (!mouse_enable) return 0;
  if (mouse_painted) {
    bitblt(mouse_x,mouse_y,mouse_width,mouse_height,mouse_screen);
  }
  getbitmap(x,y,mouse_width,mouse_height,mouse_screen);
  memcpy(mouse_paint,mouse_screen,mouse_memsize);
  miximage(mouse_paint,mouse_bitmap,mouse_width,mouse_height);
  mouse_x = x;
  mouse_y = y;
  bitblt(x,y,mouse_width,mouse_height,mouse_paint);
  mouse_painted = 1;
  mice_update  = 0;
  return 1;
}

unsigned char * bpp8_createbuffer(int w,int h)
{
  return (unsigned char *)malloc(w*h*sizeof(char));
}

int  bpp8_init(void)
{
  int n;
  /* Get Palette */
  for (n=0;n<BPP_COLOR;n++)  {
    vga_getpalette(n,&PAL[n].r,&PAL[n].g,&PAL[n].b);
    PAL[n].r <<= 2;
    PAL[n].g <<= 2;
    PAL[n].b <<= 2;
  }
  drawpixel        = vga_drawpixel;
  getpixel         = vga_getpixel;
  drawline         = vga_drawline;
  setcolor         = bpp8_setcolor;
  bitblt           = (swScanWidth==1024) ? bpp8_bitblt_1024:bpp8_bitblt;
  getbitmap        = (swScanWidth==1024) ? bpp8_getbitmap_1024:bpp8_getbitmap;
  miximage         = bpp8_miximage;
  scroll           = (swScanWidth==1024) ? bpp8_scroll_1024:bpp8_scroll;
  fillbox          = (swScanWidth==1024) ? bpp8_fillbox_1024:bpp8_fillbox;
  xorbox           = (swScanWidth==1024) ? bpp8_xorbox_1024:bpp8_xorbox;
  drawfont         = bpp8_drawfont;
  setforeground    = bpp8_setforeground;
  setbackground    = bpp8_setbackground;
  create_image_xpm = bpp8_xpm;
  mice_grinit      = bpp8_mouseinit;
  create_buffer    = bpp8_createbuffer;
  
  vga_setpalette(255,255,255,255);
  
  return 1;
}


